/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#ifndef SPEEDLINE_H
#define SPEEDLINE_H
#include <string.h>

#include <glib.h>
#include <gdk/gdk.h>
#include <gdk/gdktypes.h>
#include <glib/gstdio.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimpwidgets/gimpwidgets.h>

#include "plugin.h"

#include "speedline_intl.h"
#include "speedline_image.h"
#include "speedline_msgbox.h"

#ifdef DEBUG
#define DBGMSG(...) g_print(__VA_ARGS__)
#define WNGMSG(...) g_warning(__VA_ARGS__)
#define ERRMSG(...) g_error(__VA_ARGS__)
#else
#define DBGMSG(...) 
#define WNGMSG(...) g_warning(__VA_ARGS__)
#define ERRMSG(...) g_error(__VA_ARGS__)
#endif

typedef enum _PluginRunMode {
	RUNMODE_ADD, //mormal mode.
	RUNMODE_MODIFY //modify mode when Modify button is pushed.
} PluginRunMode;

typedef enum _AreaType {
    AREA_TYPE_RADIATION = 0,
    AREA_TYPE_PARALLEL
} AreaType;

typedef enum _LinePitchType {
	LINE_PITCH_TYPE_IN,
	LINE_PITCH_TYPE_OUT
} LinePitchType;

typedef enum _AntialiasType {
	ANTIALIAS_TYPE_NONE,
	ANTIALIAS_TYPE_FAST,
	ANTIALIAS_TYPE_BEST,
} AntialiasType;

typedef struct _DialogState {
	gint        x;
	gint        y;
	gint        width;
	gint        height;
	gboolean	topmost;
	guint		page_number;
} DialogState;

/*
 * random parameter(param_rand) has range 0.0 to 100.0. 
 * 0 has no randomize. compute next syntax.
 * param - (param * ((double)rand() / RAND_MAX + 1) * param_rand /100)
 * 
 */
typedef struct _PluginVals {
	//Line Options
	GimpRGB       line_color;
	AntialiasType line_antialias;
	gdouble       line_pitch;
	gdouble       line_pitch_rand;
	gint32        line_pitch_rand_seed;
	LinePitchType line_pitch_type;
	gdouble       line_width;
	gdouble       line_width_rand;
	gint32        line_width_rand_seed;
	gdouble       line_in_margin;
	gdouble       line_in_margin_rand;
	gint32        line_in_margin_rand_seed;
	gdouble       line_in_length;
	gdouble       line_in_length_rand;
	gint32        line_in_length_rand_seed;
	gdouble       line_out_margin;
	gdouble       line_out_margin_rand;
	gint32        line_out_margin_rand_seed;
	gdouble       line_out_length;
	gdouble       line_out_length_rand;
	gint32        line_out_length_rand_seed;

	//Area Option
	AreaType      area_type;
	GimpUnit      area_unit;

	// Radiation Area Type Options
	gdouble       radiation_x;
	gdouble       radiation_y;
	gdouble       radiation_in_radius;
	gdouble       radiation_out_radius;

	// Parallel Area Type Options
	gdouble       parallel_in_x;
	gdouble       parallel_in_y;
	gdouble       parallel_in_range;
	gdouble       parallel_out_x;
	gdouble       parallel_out_y;
	gdouble       parallel_out_range;

	//Dialog parameters. Dont save these to parasite and preset.
	gint32          image_ID;
	gint32          layer_ID;
	DialogState     gui;
	PluginRunMode   mode;
} PluginVals;

typedef struct _DialogWidgets {
	GtkWidget *dialog;
	GtkWidget *topmost;
	GtkWidget *preview;
	GtkWidget *preview_plus_button;
	GtkWidget *preview_minus_button;

	GtkWidget *line_antialias;
	GtkWidget *line_pitch;
	GtkWidget *line_pitch_rand;
	GtkWidget *line_pitch_rand_seed;
	GtkWidget *line_pitch_type;
	GtkWidget *line_width;
	GtkWidget *line_width_rand;
	GtkWidget *line_width_rand_seed;
	GtkWidget *line_in_margin;
	GtkWidget *line_in_margin_rand;
	GtkWidget *line_in_margin_rand_seed;
	GtkWidget *line_out_margin;
	GtkWidget *line_out_margin_rand;
	GtkWidget *line_out_margin_rand_seed;
	GtkWidget *line_in_length;
	GtkWidget *line_in_length_rand;
	GtkWidget *line_in_length_rand_seed;
	GtkWidget *line_out_length;
	GtkWidget *line_out_length_rand;
	GtkWidget *line_out_length_rand_seed;

	GtkWidget *area_type_unit;
	GtkWidget *area_type_radiation;
	GtkWidget *area_type_parallel;

	GtkWidget *area_type_notebook;

	GtkWidget *radiation_x;
	GtkWidget *radiation_y;
	GtkWidget *radiation_in_radius;
	GtkWidget *radiation_out_radius;

	GtkWidget *parallel_in_x;
	GtkWidget *parallel_in_y;
	GtkWidget *parallel_in_range;
	GtkWidget *parallel_out_x;
	GtkWidget *parallel_out_y;
	GtkWidget *parallel_out_range;

	GtkWidget		*preset_frame;
	GtkWidget		*preset_listbox;
	GtkListStore	*preset_store;
	GtkWidget		*preset_tree;
	GtkWidget		*preset_load;
	GtkWidget		*preset_save;
	GtkWidget		*preset_delete;
	GtkWidget		*preset_update;

} DialogWidgets;

extern DialogWidgets widgets;

#endif
/* vim: set fdm=marker ts=4 sw=4:*/
