/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline_preview.h"
#include "speedline_parallel.h"
#include "speedline_line.h"

typedef enum _ParallelGuideObj {
	PARALLEL_GUIDE_OBJ_NONE = 0,
	PARALLEL_GUIDE_OBJ_IN_POINT,
	PARALLEL_GUIDE_OBJ_OUT_POINT,
	PARALLEL_GUIDE_OBJ_IN_LINE,
	PARALLEL_GUIDE_OBJ_OUT_LINE,
	PARALLEL_GUIDE_OBJ_IN_LENGTH_POINT,
	PARALLEL_GUIDE_OBJ_OUT_LENGTH_POINT,
} ParallelGuideObj;

static void path_guide_in_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_out_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_in_line (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_out_line (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_in_length_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_out_length_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_connected_length_line (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void get_global_vals (PluginVals *vals, gdouble	*in_x, gdouble	*in_y, gdouble	*in_range, gdouble	*out_x, gdouble	*out_y, gdouble	*out_range);
static ParallelGuideObj parallel_guide_hit_test (cairo_t *cr, GdkWindow *window, PluginVals	*vals, gdouble x, gdouble y);


static void
path_guide_in_point (cairo_t *cr,
					 gdouble in_x,
					 gdouble in_y,
					 gdouble in_range,
					 gdouble out_x,
					 gdouble out_y,
					 gdouble out_range)
{
	double size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, in_x, in_y, preview_local_to_global (size), 0, 2 * G_PI);
}

static void
path_guide_out_point (cairo_t *cr,
					  gdouble in_x,
					  gdouble in_y,
					  gdouble in_range,
					  gdouble out_x,
					  gdouble out_y,
					  gdouble out_range)
{
	double size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, out_x, out_y, preview_local_to_global (size), 0, 2 * G_PI);
}

static void
path_guide_in_line (cairo_t *cr,
					gdouble in_x,
					gdouble in_y,
					gdouble in_range,
					gdouble out_x,
					gdouble out_y,
					gdouble out_range)
{
	cairo_save (cr);
	cairo_translate (cr, in_x, in_y);
	cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
	cairo_move_to (cr, in_range / -2.0, 0);
	cairo_line_to (cr, in_range /  2.0, 0);
	cairo_restore (cr);
}

static void
path_guide_out_line (cairo_t *cr,
					 gdouble in_x,
					 gdouble in_y,
					 gdouble in_range,
					 gdouble out_x,
					 gdouble out_y,
					 gdouble out_range)
{
	cairo_save (cr);
	cairo_translate (cr, out_x, out_y);
	cairo_rotate (cr, G_PI_2 + atan2 (in_y - out_y, in_x - out_x));
	cairo_move_to (cr, out_range / -2.0, 0);
	cairo_line_to (cr, out_range /  2.0, 0);
	cairo_restore (cr);
}

static void
path_guide_in_length_point (cairo_t *cr,
							gdouble in_x,
							gdouble in_y,
							gdouble in_range,
							gdouble out_x,
							gdouble out_y,
							gdouble out_range)
{
	double size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;

	cairo_save (cr);
	cairo_translate (cr, in_x, in_y);
	cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
	cairo_arc (cr, in_range / -2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_arc (cr, in_range /  2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_restore (cr);
}

static void
path_guide_out_length_point (cairo_t *cr,
							 gdouble in_x,
							 gdouble in_y,
							 gdouble in_range,
							 gdouble out_x,
							 gdouble out_y,
							 gdouble out_range)
{
	double size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;

	cairo_save (cr);
	cairo_translate (cr, out_x, out_y);
	cairo_rotate (cr, G_PI_2 + atan2 (in_y - out_y, in_x - out_x));
	cairo_arc (cr, out_range / -2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_arc (cr, out_range /  2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_restore (cr);
}

static void
path_guide_connected_length_line (cairo_t *cr,
								  gdouble in_x,
								  gdouble in_y,
								  gdouble in_range,
								  gdouble out_x,
								  gdouble out_y,
								  gdouble out_range)
{
	gdouble len;

	len = sqrt ((out_x - in_x) * (out_x - in_x) + (out_y - in_y) * (out_y - in_y));

	cairo_save (cr);
	cairo_translate (cr, in_x, in_y);
	cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
	cairo_move_to (cr,  in_range / -2.0, 0);
	cairo_line_to (cr, out_range / -2.0, -len);
	cairo_move_to (cr,  in_range /  2.0, 0);
	cairo_line_to (cr, out_range /  2.0, -len);
	cairo_restore (cr);
}

static void 
get_global_vals (PluginVals *vals,
				 gdouble	*in_x,
				 gdouble	*in_y,
				 gdouble	*in_range,
				 gdouble	*out_x,
				 gdouble	*out_y,
				 gdouble	*out_range)
{
	gdouble resx, resy;
	gimp_image_get_resolution (vals->image_ID, &resx, &resy);

	*in_x = gimp_units_to_pixels (vals->parallel_in_x, vals->area_unit, resx);
	*in_y = gimp_units_to_pixels (vals->parallel_in_y, vals->area_unit, resx);
	*in_range = gimp_units_to_pixels (vals->parallel_in_range, vals->area_unit, resx);
	*out_x = gimp_units_to_pixels (vals->parallel_out_x, vals->area_unit, resx);
	*out_y = gimp_units_to_pixels (vals->parallel_out_y, vals->area_unit, resx);
	*out_range = gimp_units_to_pixels (vals->parallel_out_range, vals->area_unit, resx);
}

static ParallelGuideObj
parallel_guide_hit_test (cairo_t	*cr,
						 GdkWindow	*window,
						 PluginVals	*vals,
						 gdouble	x,
						 gdouble	y)
{
	gdouble in_x, in_y, in_range, out_x, out_y, out_range;

	get_global_vals (vals, &in_x, &in_y, &in_range, &out_x, &out_y, &out_range);

	cairo_set_source_rgba (cr, 1, 1, 1, 1);
	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH_FOR_HITTEST));
	cairo_device_to_user (cr, &x, &y);

	//check if in_point
	path_guide_in_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_IN_POINT;
	}
	cairo_new_path (cr);

	//check if out_point
	path_guide_out_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_OUT_POINT;
	}
	cairo_new_path (cr);
	
	//check if in_line
	path_guide_in_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_stroke (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_IN_LINE;
	}
	cairo_new_path (cr);

	//check if out_line
	path_guide_out_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_stroke (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_OUT_LINE;
	}
	cairo_new_path (cr);

	//check if in_length_point
	path_guide_in_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_IN_LENGTH_POINT;
	}
	cairo_new_path (cr);

	//check if out_length_point
	path_guide_out_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_OUT_LENGTH_POINT;
	}
	cairo_new_path (cr);

	return PARALLEL_GUIDE_OBJ_NONE;
}

void
parallel_draw_guide (cairo_t *cr, PluginVals *vals)
{
	gdouble in_x, in_y, in_range, out_x, out_y, out_range;

	get_global_vals (vals, &in_x, &in_y, &in_range, &out_x, &out_y, &out_range);

	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH));
	cairo_set_source_rgba (cr,
		PREVIEW_GUIDE_COLOR_R,
		PREVIEW_GUIDE_COLOR_G,
		PREVIEW_GUIDE_COLOR_B,
		1);


	//in point
	path_guide_in_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//out point
	path_guide_out_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//in line
	path_guide_in_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_stroke (cr);

	//out line
	path_guide_out_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_stroke (cr);

	//in length point
	path_guide_in_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//out length point
	path_guide_out_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//connected length line
	path_guide_connected_length_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_stroke (cr);


	//lines
	{
		int n, nmax;
		gdouble ylen;
		ylen = sqrt ((out_x - in_x) * (out_x - in_x) + (out_y - in_y) * (out_y - in_y));


		cairo_save (cr);
		cairo_translate (cr, in_x, in_y);
		cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
		nmax = PREVIEW_GUIDE_LINE_NUMBER;
		for (n = 0; n < nmax; n++) {
			cairo_move_to (cr,  (in_range  / -2.0) + in_range  * ((double)n/nmax), 0);
			cairo_line_to (cr,  (out_range / -2.0) + out_range * ((double)n/nmax), -ylen);
		}
		cairo_restore (cr);
		cairo_stroke (cr);
	}
}

void
parallel_draw_line (cairo_t *cr, PluginVals *vals)
{
	LineObj *line = line_new (vals);
	gdouble in_x, in_y, in_range, out_x, out_y, out_range;

	get_global_vals (vals, &in_x, &in_y, &in_range, &out_x, &out_y, &out_range);

	cairo_save (cr);
	{
		double p1, psum;
		double max_len = 0;
		double distance_in_out = 0;

		distance_in_out = sqrt ((in_y - out_y) * (in_y - out_y) + (in_x - out_x) * (in_x - out_x));

		//compute length for line_pitch
		switch (vals->line_pitch_type) {
			case LINE_PITCH_TYPE_IN:  max_len = in_range; break;
			case LINE_PITCH_TYPE_OUT: max_len = out_range; break;
		}
		p1 = line->pitch / max_len;
		psum = 0;

		g_assert (p1 != 0);

		//normalize area center and angle
		cairo_translate (cr, in_x, in_y);
		cairo_rotate (cr, atan2 ((out_y - in_y), (out_x - in_x)));
		cairo_rotate (cr, -G_PI_2);

		while (psum <= in_range) {
			double p = p1 * in_range;
			double inx, iny, outx, outy;

			p = p - (p * g_rand_double (line->randobj_pitch) * line->pitch_rand / 100.0);
			inx = -in_range / 2.0 + psum;
			iny = 0;
			outx = -out_range / 2.0 + out_range * (psum / in_range);
			outy = distance_in_out;

			cairo_save (cr);
			{
				//normaze line center and angle
				double line_length;
				line_length = sqrt ((iny - outy) * (iny - outy) + (inx - outx) * (inx - outx));

				cairo_translate (cr, inx, iny);
				cairo_rotate (cr, atan2 (outy - iny, outx - inx));
				cairo_rotate (cr, -G_PI_2);

				line_draw (line, cr, line_length);
			}
			cairo_restore (cr);
			psum += p;
		}
	}
	cairo_restore (cr);

	line_destroy (line);
}

gboolean
parallel_event_hook (cairo_t *cr, GdkEvent *event, PluginVals *vals)
{
	static ParallelGuideObj dragging_obj = PARALLEL_GUIDE_OBJ_NONE;
	static gdouble          dragging_start_point_x = 0;
	static gdouble          dragging_start_point_y = 0;
	static gdouble          dragging_start_in_x = 0;
	static gdouble          dragging_start_in_y = 0;
	static gdouble          dragging_start_out_x = 0;
	static gdouble          dragging_start_out_y = 0;

	switch (event->type) {
		case GDK_EXPOSE:
			break;

		case GDK_BUTTON_PRESS:
			dragging_start_point_x = event->button.x;
			dragging_start_point_y = event->button.y;
			dragging_start_in_x = vals->parallel_in_x;
			dragging_start_in_y = vals->parallel_in_y;
			dragging_start_out_x = vals->parallel_out_x;
			dragging_start_out_y = vals->parallel_out_y;
			dragging_obj = parallel_guide_hit_test (cr, event->any.window, vals, event->button.x, event->button.y);
			break;

		case GDK_BUTTON_RELEASE:
			dragging_start_point_x = 0;
			dragging_start_point_y = 0;
			dragging_start_in_x = 0;
			dragging_start_in_y = 0;
			dragging_start_out_x = 0;
			dragging_start_out_y = 0;
			dragging_obj = PARALLEL_GUIDE_OBJ_NONE;
			break;

		case GDK_MOTION_NOTIFY:
			if (dragging_obj != PARALLEL_GUIDE_OBJ_NONE) {
				gdouble gx, gy;
				gdouble resx, resy;
				gimp_image_get_resolution (vals->image_ID, &resx, &resy);
				gx = event->motion.x;
				gy = event->motion.y;
				cairo_device_to_user (cr, &gx, &gy);
				
				switch (dragging_obj) {
					case PARALLEL_GUIDE_OBJ_IN_POINT:
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_x), gimp_pixels_to_units (gx, vals->area_unit, resx));
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_y), gimp_pixels_to_units (gy, vals->area_unit, resy));
						break;
					case PARALLEL_GUIDE_OBJ_OUT_POINT:
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_x), gimp_pixels_to_units (gx, vals->area_unit, resx));
						gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_y), gimp_pixels_to_units (gy, vals->area_unit, resy));
						break;

					case PARALLEL_GUIDE_OBJ_IN_LINE:
						{
							gdouble valx, valy;

							if (dragging_start_in_x == dragging_start_out_x) {
								valx = dragging_start_in_x;
								valy = gy;
							} else if (dragging_start_in_y == dragging_start_out_y) {
								valx = gx;
								valy = dragging_start_in_y;
							} else {
								gdouble nx, ny;
								gdouble m1, n1;// y = (m1*x)+n1;
								gdouble m2, n2;// y = (m2*x)+n2;
								nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
								ny = gimp_pixels_to_units (gy, vals->area_unit, resy);

								m1 = (dragging_start_out_y - dragging_start_in_y) / (dragging_start_out_x - dragging_start_in_x);
								n1 = dragging_start_in_y - (m1 * dragging_start_in_x);
								m2 = -1.0 / m1;
								n2 = ny - m2 * nx;

								valx = (n2 - n1) / (m1 - m2);
								valy = valx * m1 + n1;
							}
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_x), valx);
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_y), valy);
						}
						break;
					case PARALLEL_GUIDE_OBJ_OUT_LINE:
						{
							gdouble valx, valy;

							if (dragging_start_in_x == dragging_start_out_x) {
								valx = dragging_start_in_x;
								valy = gy;
							} else if (dragging_start_in_y == dragging_start_out_y) {
								valx = gx;
								valy = dragging_start_in_y;
							} else {
								gdouble nx, ny;
								gdouble m1, n1;// y = (m1*x)+n1;
								gdouble m2, n2;// y = (m2*x)+n2;
								nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
								ny = gimp_pixels_to_units (gy, vals->area_unit, resy);

								m1 = (dragging_start_in_y - dragging_start_out_y) / (dragging_start_in_x - dragging_start_out_x);
								n1 = dragging_start_out_y - (m1 * dragging_start_out_x);
								m2 = -1.0 / m1;
								n2 = ny - m2 * nx;

								valx = (n2 - n1) / (m1 - m2);
								valy = valx * m1 + n1;
							}
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_x), valx);
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_y), valy);
						}
						break;

					case PARALLEL_GUIDE_OBJ_IN_LENGTH_POINT:
						{
							gdouble px, py, nx, ny, val;

							px = gimp_pixels_to_units (vals->parallel_in_x, vals->area_unit, resx);
							py = gimp_pixels_to_units (vals->parallel_in_y, vals->area_unit, resy);
							nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
							ny = gimp_pixels_to_units (gy, vals->area_unit, resy);
							val = sqrt ((ny - py) * (ny - py) + (nx - px) * (nx - px)) * 2;
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_range), val);
						}
						break;
					case PARALLEL_GUIDE_OBJ_OUT_LENGTH_POINT:
						{
							gdouble px, py, nx, ny, val;

							px = gimp_pixels_to_units (vals->parallel_out_x, vals->area_unit, resx);
							py = gimp_pixels_to_units (vals->parallel_out_y, vals->area_unit, resy);
							nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
							ny = gimp_pixels_to_units (gy, vals->area_unit, resy);
							val = sqrt ((ny - py) * (ny - py) + (nx - px) * (nx - px)) * 2;
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_range), val);
						}
						break;
				}
				gtk_widget_queue_draw (widgets.preview);
			}
			break;

		case GDK_KEY_PRESS:
			break;

		case GDK_KEY_RELEASE:
			break;

		default:
			break;
	}
	return FALSE;
}

/* vim: set fdm=marker ts=4 sw=4:*/
